@@ -41,6 +41,7 @@
 #include "xslt.h"
 #include "xsltInternals.h"
 #include "xsltutils.h"
+#include "preproc.h"
 #include "imports.h"
 #include "documents.h"
 #include "security.h"
@@ -58,14 +59,13 @@
  * @style: the stylesheet being imported by the master
  *
  * normalize the comp steps for the stylesheet being imported
- * by the master, together with any imports within that. 
+ * by the master, together with any imports within that.
  *
  */
-static void xsltFixImportedCompSteps(xsltStylesheetPtr master, 
+static void xsltFixImportedCompSteps(xsltStylesheetPtr master,
 			xsltStylesheetPtr style) {
     xsltStylesheetPtr res;
-    xmlHashScan(style->templatesHash,
-	            (xmlHashScanner) xsltNormalizeCompSteps, master);
+    xmlHashScan(style->templatesHash, xsltNormalizeCompSteps, master);
     master->extrasNr += style->extrasNr;
     for (res = style->imports; res != NULL; res = res->next) {
         xsltFixImportedCompSteps(master, res);
@@ -95,7 +95,7 @@ xsltParseStylesheetImport(xsltStylesheetPtr style, xmlNodePtr cur) {
     if ((cur == NULL) || (style == NULL))
 	return (ret);
 
-    uriRef = xsltGetNsProp(cur, (const xmlChar *)"href", XSLT_NAMESPACE);
+    uriRef = xmlGetNsProp(cur, (const xmlChar *)"href", NULL);
     if (uriRef == NULL) {
 	xsltTransformError(NULL, style, cur,
 	    "xsl:import : missing href attribute\n");
@@ -189,11 +189,12 @@ xsltParseStylesheetInclude(xsltStylesheetPtr style, xmlNodePtr cur) {
     xsltStylesheetPtr result;
     xsltDocumentPtr include;
     xsltDocumentPtr docptr;
+    int oldNopreproc;
 
     if ((cur == NULL) || (style == NULL))
 	return (ret);
 
-    uriRef = xsltGetNsProp(cur, (const xmlChar *)"href", XSLT_NAMESPACE);
+    uriRef = xmlGetNsProp(cur, (const xmlChar *)"href", NULL);
     if (uriRef == NULL) {
 	xsltTransformError(NULL, style, cur,
 	    "xsl:include : missing href attribute\n");
@@ -228,13 +229,32 @@ xsltParseStylesheetInclude(xsltStylesheetPtr style, xmlNodePtr cur) {
 	    "xsl:include : unable to load %s\n", URI);
 	goto error;
     }
-
+#ifdef XSLT_REFACTORED
+    if (IS_XSLT_ELEM_FAST(cur) && (cur->psvi != NULL)) {
+	((xsltStyleItemIncludePtr) cur->psvi)->include = include;
+    } else {
+	xsltTransformError(NULL, style, cur,
+	    "Internal error: (xsltParseStylesheetInclude) "
+	    "The xsl:include element was not compiled.\n", URI);
+	style->errors++;
+    }
+#endif
     oldDoc = style->doc;
     style->doc = include->doc;
     /* chain to stylesheet for recursion checking */
     include->includes = style->includes;
     style->includes = include;
+    oldNopreproc = style->nopreproc;
+    style->nopreproc = include->preproc;
+    /*
+    * TODO: This will change some values of the
+    *  including stylesheet with every included module
+    *  (e.g. excluded-result-prefixes)
+    *  We need to strictly seperate such stylesheet-owned values.
+    */
     result = xsltParseStylesheetProcess(style, include->doc);
+    style->nopreproc = oldNopreproc;
+    include->preproc = 1;
     style->includes = include->includes;
     style->doc = oldDoc;
     if (result == NULL) {
@@ -273,7 +293,7 @@ xsltNextImport(xsltStylesheetPtr cur) {
 	return(cur->next) ;
     do {
 	cur = cur->parent;
-	if (cur == NULL) return(NULL);
+	if (cur == NULL) break;
 	if (cur->next != NULL) return(cur->next);
     } while (cur != NULL);
     return(cur);
@@ -308,7 +328,7 @@ xsltNeedElemSpaceHandling(xsltTransformContextPtr ctxt) {
  * @ctxt:  an XSLT transformation context
  * @node:  an XML node
  *
- * Find strip-space or preserve-space informations for an element
+ * Find strip-space or preserve-space information for an element
  * respect the import precedence or the wildcards
  *
  * Returns 1 if space should be stripped, 0 if not, and 2 if everything
@@ -327,6 +347,11 @@ xsltFindElemSpaceHandling(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	if (node->ns != NULL) {
 	    val = (const xmlChar *)
 	      xmlHashLookup2(style->stripSpaces, node->name, node->ns->href);
+            if (val == NULL) {
+                val = (const xmlChar *)
+                    xmlHashLookup2(style->stripSpaces, BAD_CAST "*",
+                                   node->ns->href);
+            }
 	} else {
 	    val = (const xmlChar *)
 		  xmlHashLookup2(style->stripSpaces, node->name, NULL);
@@ -336,7 +361,7 @@ xsltFindElemSpaceHandling(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		return(1);
 	    if (xmlStrEqual(val, (xmlChar *) "preserve"))
 		return(0);
-	} 
+	}
 	if (style->stripAll == 1)
 	    return(1);
 	if (style->stripAll == -1)
@@ -354,6 +379,13 @@ xsltFindElemSpaceHandling(xsltTransformContextPtr ctxt, xmlNodePtr node) {
  * @nameURI: the template name URI
  *
  * Finds the named template, apply import precedence rule.
+ * REVISIT TODO: We'll change the nameURI fields of
+ *  templates to be in the string dict, so if the
+ *  specified @nameURI is in the same dict, then use pointer
+ *  comparison. Check if this can be done in a sane way.
+ *  Maybe this function is not needed internally at
+ *  transformation-time if we hard-wire the called templates
+ *  to the caller.
  *
  * Returns the xsltTemplatePtr or NULL if not found
  */
@@ -367,17 +399,12 @@ xsltFindTemplate(xsltTransformContextPtr ctxt, const xmlChar *name,
 	return(NULL);
     style = ctxt->style;
     while (style != NULL) {
-	cur = style->templates;
-	while (cur != NULL) {
-	    if (xmlStrEqual(name, cur->name)) {
-		if (((nameURI == NULL) && (cur->nameURI == NULL)) ||
-		    ((nameURI != NULL) && (cur->nameURI != NULL) &&
-		     (xmlStrEqual(nameURI, cur->nameURI)))) {
-		    return(cur);
-		}
-	    }
-	    cur = cur->next;
-	}
+        if (style->namedTemplates != NULL) {
+            cur = (xsltTemplatePtr)
+                xmlHashLookup2(style->namedTemplates, name, nameURI);
+            if (cur != NULL)
+                return(cur);
+        }
 
 	style = xsltNextImport(style);
     }
